// *****************************************************************
// Filename:    AspectGraphProcessor.cpp
// Copyright:   Kai Welke, Chair Prof. Dillmann (IAIM),
//              Institute for Computer Science and Engineering (CSE),
//              University of Karlsruhe. All rights reserved.
// Author:      Kai Welke
// Date:        12.06.2007
// *****************************************************************

// *****************************************************************
// includes
// *****************************************************************
#include "MathTools.h"
#include <math.h>
//#include "Base/Tools/DebugMemory.h"


// **********************************************************
// transformation on sphere coordinates and paths
// **********************************************************
// returns transformation from p1 -> p2
TTransform MathTools::determinePathTransformation(TPath* p1, TPath* p2)
{
	TSphereCoord start1, start2, end1, end2;
	start1 = p1->at(0).Position;
	end1 = p1->at(p1->size()-1).Position;
	start2 = p2->at(0).Position;
	end2 = p2->at(p2->size()-1).Position;

	return determineTransformation(start1,start2,end1,end2);
}

TPath* MathTools::transformPath(TTransform transform, TPath* pPath)
{
	TPath* pResulting = new TPath();

	if(pPath->size() < 2)
	{
		printf("ERROR: path is too short\n");
		delete pResulting;
		return NULL;
	}
	
	// copy path and transform
	for(int e = 0 ; e < int(pPath->size()) ; e++)
	{
		TPathElement element = pPath->at(e);
		TSphereCoord res = MathTools::transform(element.Position, transform);
		element.Position = res;

		pResulting->push_back(element);
	}

	return pResulting;

}

TPath* MathTools::inverseTransformPath(TTransform transform, TPath* pPath)
{
	TPath* pResulting = new TPath();

	if(pPath->size() < 2)
	{
		printf("ERROR: path is too short\n");
		delete pResulting;
		return NULL;
	}
	
	// copy path and transform
	for(int e = 0 ; e < int(pPath->size()) ; e++)
	{
		TPathElement element = pPath->at(e);
		TSphereCoord res = MathTools::inverseTransform(element.Position, transform);
		element.Position = res;

		pResulting->push_back(element);
	}

	return pResulting;

}

TTransform MathTools::determineTransformation(TSphereCoord start1, TSphereCoord start2, TSphereCoord end1, TSphereCoord end2)
{
	TTransform result;

	// first determine alpha and beta and beta axis
	float fAlpha, fBeta;
	Vec3d betaAxis;
	determineTransformationAlphaBeta(start1,start2,fAlpha,fBeta,betaAxis);
	result.fAlpha = fAlpha;
	result.fBeta = fBeta;
	result.betaAxis = betaAxis;

	// calculate angular for psi rotation
	
	// apply transform to first path end
	TSphereCoord end1_transformed = transformAlphaBeta(end1, fAlpha, fBeta, betaAxis);

	Vec3d end1_transformed_v, start2_v, end2_v;
	convert(end1_transformed, end1_transformed_v);
	convert(start2, start2_v);
	convert(end2, end2_v);

	// calculate angle between start2 <-> end1_tranformed ; start2 <-> end2 
	Vec3d zero;
	Math3d::SetVec(zero,0,0,0);

	Vec3d line1 = MathTools::dropAPerpendicular(end1_transformed_v, zero, start2_v);
	Vec3d line2 = MathTools::dropAPerpendicular(end2_v, zero, start2_v);
	Math3d::SubtractFromVec(line1, end1_transformed_v);
	Math3d::SubtractFromVec(line2, end2_v);

	float fPsi = MathTools::AngleAndDirection(line1,line2,start2_v) * 180.0 / M_PI;

	result.fPsi = fPsi;
	result.psiAxis = start2_v;

	return result;
}

TSphereCoord MathTools::transform(TSphereCoord sc, TTransform transform)
{
	// first apply alpha , beta
	TSphereCoord rotated_ab = transformAlphaBeta(sc, transform.fAlpha, transform.fBeta, transform.betaAxis);

	Vec3d rotated_ab_v;

	convert(rotated_ab, rotated_ab_v);

	// now apply final rotation
	Mat3d rot;
	Math3d::SetRotationMatAxis(rot, transform.psiAxis, -transform.fPsi * M_PI / 180.0);
	Math3d::MulMatVec(rot, rotated_ab_v, rotated_ab_v);

	TSphereCoord rotated;
	convert(rotated_ab_v, rotated);

	return rotated;
}

TSphereCoord MathTools::inverseTransform(TSphereCoord sc, TTransform transform)
{
	Vec3d sc_v;
	convert(sc, sc_v);

	// first rotate
	Mat3d rot;
	Math3d::SetRotationMatAxis(rot, transform.psiAxis, transform.fPsi * M_PI / 180.0);
	Math3d::MulMatVec(rot, sc_v, sc_v);

	TSphereCoord rotated;
	convert(sc_v, rotated);

	return inverseTransformAlphaBeta(rotated, transform.fAlpha, transform.fBeta, transform.betaAxis	);
}

// determine alpha and beta of a transform
// fAlpha: rotation around the z axis in counterclockwise direction considering the positive z-axis
// fBeta: rotation around the rotated y axis in counterclockwise direction considering the positive y-axis
void MathTools::determineTransformationAlphaBeta(TSphereCoord sc1, TSphereCoord sc2, float& fAlpha, float& fBeta, Vec3d& betaAxis)
{
	Vec3d vector1, vector2;
	convert(sc1, vector1);
	convert(sc2, vector2);

	// project both vectors to xy plane
	Vec3d vector1_xy, vector2_xy;

	vector1_xy = vector1;
	vector2_xy = vector2;
	vector1_xy.z = 0.0;
	vector2_xy.z = 0.0;

	// calculate first angle from vec1_projected to vec2_projected with normale z
	Vec3d normal;
	Math3d::SetVec(normal,0.0,0.0,1.0);

	if( (Math3d::Length(vector1_xy) == 0.0) || (Math3d::Length(vector2_xy) == 0))
		fAlpha = (( sc2.fTheta - sc1.fTheta ) / 180.0 * M_PI);
	else
		fAlpha = MathTools::AngleAndDirection(vector1_xy, vector2_xy, normal);
	
	// now rotate first vector with alpha
	Vec3d rotated1,rotation, rotated_y, rotation_yaxis;

	// the vector to rotate
	Math3d::SetVec(rotated1, vector1);

	// set rotation: alpha around z-axis to transform v1 in same plane as v2
	Math3d::SetVec(rotation, 0.0,0.0,fAlpha);
	Math3d::SetVec(rotation_yaxis, 0.0,0.0, sc2.fTheta / 180.0 * M_PI);

	// we need normal for angle calculateion, so also rotate x-axis
	Math3d::SetVec(rotated_y,0.0,1.0,0.0);

	// rotate vector and x-axis
	Math3d::RotateVec(rotated1, rotation, rotated1);
	Math3d::RotateVec(rotated_y, rotation_yaxis, rotated_y);

	betaAxis = rotated_y;

	// calculate angle between both vectors
	fBeta = MathTools::AngleAndDirection(rotated1, vector2, rotated_y);

	fAlpha *= 180.0 / M_PI;
	fBeta *= 180.0 / M_PI;
}

TSphereCoord MathTools::transformAlphaBeta(TSphereCoord sc, float fAlpha, float fBeta, Vec3d betaAxis)
{
	TSphereCoord result;
    Vec3d vector1, vector1_rotated;
	convert(sc, vector1);

	// rotate around positive z-axis
	Vec3d rotation_alpha;
	Math3d::SetVec(rotation_alpha, 0.0,0.0, fAlpha / 180.0 * M_PI);

	// rotate vector aroun z axis
	Math3d::RotateVec(vector1,rotation_alpha, vector1_rotated);

	// rotate vector with beta around betaAxis
	Mat3d rot;
	Math3d::SetRotationMatAxis(rot, betaAxis, -fBeta / 180.0 * M_PI);
	Math3d::MulMatVec(rot,vector1_rotated,vector1_rotated);

	convert(vector1_rotated,result);

	return result;
}

TSphereCoord MathTools::inverseTransformAlphaBeta(TSphereCoord sc, float fAlpha, float fBeta, Vec3d betaAxis)
{
	Vec3d vector1;
	convert(sc, vector1);

	// do inverse of beta axis rotation
	Mat3d rot;
	Math3d::SetRotationMatAxis(rot, betaAxis, fBeta / 180.0 * M_PI);
	Math3d::MulMatVec(rot,vector1,vector1);

	// do inverse of zaxis rotation
	Vec3d rotation_alpha;
	Math3d::SetVec(rotation_alpha, 0.0,0.0, -fAlpha / 180.0 * M_PI);

	// rotate vector aroun z axis
	Math3d::RotateVec(vector1,rotation_alpha, vector1);

	TSphereCoord result;
	convert(vector1, result);

	return result;
}

// **********************************************************
// calculations on sphere coordinates
// **********************************************************
float MathTools::CalculateCrossProductFromDifference(TSphereCoord p1, TSphereCoord p2, TSphereCoord p3)
{
	float u1, v1, u2, v2;

	u1 =  p2.fPhi - p1.fPhi;
	v1 =  p2.fTheta - p1.fTheta;
	u2 =  p3.fPhi - p1.fPhi;
	v2 =  p3.fTheta - p1.fTheta;

	return u1 * v2 - v1 * u2;
}

float MathTools::getDistanceOnArc(TSphereCoord p1, TSphereCoord p2)
{
	Vec3d vec1, vec2;
	p1.fDistance = 1.0f;
	p2.fDistance = 1.0f;

	convert(p1,vec1);
	convert(p2,vec2);

	float fAngle = (float) Math3d::Angle(vec1,vec2);

	// length of circular arc
	// not necessary because Math3d::Angle returns a value in radians
	//float fDist = float( (2 * M_PI * fAngle ) / 360.0f);

	return fAngle;
}

float MathTools::getDistance(TSphereCoord p1, TSphereCoord p2, float fRadius)
{
	// never use this!!!!
	TSphereCoord diff;
	diff.fPhi = p1.fPhi - p2.fPhi;
	diff.fTheta = p1.fTheta - p2.fTheta;

	float fAngle = sqrt(diff.fPhi * diff.fPhi + diff.fTheta * diff.fTheta);

	return fAngle;
}

// **********************************************************
// extended 2d Math
// **********************************************************
bool MathTools::IntersectLines(Vec2d p1, Vec2d m1, Vec2d p2, Vec2d m2, Vec2d& res)
{
	// calculate perpendicular bisector of sides
	float dx = float(p1.x - p2.x);
	float dy = float(p1.y - p2.y);

	// now from
	// g1: g1(x,y) = (p1x,p1y) + (m1x,m1y) * c1
	// g2: g2(x,y) = (p2x,p2y) + (m2x,m2y) * c2
	// -->
	// g1(x,y) = g2(x,y)
	// -->
	// (p1x,p1y) + (m1x,m1y) * c1 = (p2x,p2y) + (m2x,m2y) * c2
	// (p1x,p1y) - (p2x,p2y) = (dx,dy) = (m2x,m2y) * c2 - (m1x,m1y) * c1
	// -->
	// dx = m2x * c2 - m1x * c1,  dy = m2y * c2 - m1y * c1
	// c1 = (m2y * c2 - dy) / m1y
	// dx = m2x * c2 - m1x * (m2y * c2 - dy) / m1y
	
	float z = float(dx - (m1.x * dy) / m1.y);
	float n = float(m2.x - (m1.x * m2.y) / m1.y);

	float c2;
	
	if(n==0)
		return false;
	else
		c2 = z / n;

	res.x = p2.x + c2 * m2.x;
	res.y = p2.y + c2 * m2.y;

	return true;
}

// **********************************************************
// extended 3d Math
// **********************************************************
// checked and working
float MathTools::AngleAndDirection(Vec3d vector1, Vec3d vector2, Vec3d normal)
{
	double fAngle = Math3d::Angle(vector1, vector2);

	Vec3d axis;
	Math3d::CrossProduct(vector1, vector2, axis);

	// compare axis and normal
	Math3d::NormalizeVec(normal);
	double c = Math3d::ScalarProduct(normal,vector1);

	Math3d::AddToVec(axis, vector1);
	double d = Math3d::ScalarProduct(normal, axis) - c;


	if( d < 0)
		return 2* M_PI - fAngle;

	return (float) fAngle;
}

// checked and working
Vec3d MathTools::dropAPerpendicular(Vec3d point, Vec3d linepoint, Vec3d linevector)
{
	// helping plane normal (ax + by + cz + d = 0)
	Vec3d plane_normal = linevector; // (a,b,c)
	float d = (float) -Math3d::ScalarProduct(plane_normal, point);

	// calculate k from plane and line intersection
	float z = float( - plane_normal.x * linepoint.x - plane_normal.y * linepoint.y - plane_normal.z * linepoint.z - d);
	float n = float( plane_normal.x * linevector.x + plane_normal.y * linevector.y + plane_normal.z * linevector.z);

	float k = z / n;

	// calculate line from line equation
	Math3d::MulVecScalar(linevector,k,linevector);
	Math3d::AddToVec(linevector, linepoint);

	return linevector;

}

// **********************************************************
// conversions
// **********************************************************
// checked and working
void MathTools::convert(TSphereCoord in, Vec3d& out)
{
	// alpha is azimuth, beta is polar angle
	in.fPhi *= M_PI / 180.0f;
	in.fTheta *= M_PI / 180.0f;

// OLD and wrong??
//	out.x = in.fDistance * sin(in.fPhi) * cos(in.fTheta);
//	out.y = in.fDistance * sin(in.fPhi) * sin(in.fTheta);
//	out.z = in.fDistance * cos(in.fPhi); 

	Vec3d vector;
	Vec3d rotation_y, rotation_z;

	Math3d::SetVec(vector,0.0,0.0,1.0);

	Math3d::SetVec(rotation_y, 0.0, in.fPhi, 0.0);
	Math3d::SetVec(rotation_z, 0.0, 0.0, in.fTheta);

	// first rotate around y axis with phi
	Math3d::RotateVec(vector, rotation_y, vector);

	// second rotate around z with theta
	Math3d::RotateVec(vector, rotation_z, vector);
	out = vector;
}

void MathTools::convert(TSphereCoord in, Eigen::Vector3f& out)
{
    Vec3d vec;
    convert(in, vec);
    out(0) = vec.x;
    out(1) = vec.y;
    out(2) = vec.z;
}

void MathTools::convertWithDistance(TSphereCoord in, Vec3d& out)
{
  	convert(in, out);
	Math3d::MulVecScalar(out, in.fDistance, out);
}

// checked and working
void MathTools::convert(Vec3d in, TSphereCoord& out)
{
	// alpha is azimuth, beta is polar angle
	out.fDistance = (float) sqrt(in.x * in.x + in.y * in.y + in.z * in.z);

	float fAngle = in.z / out.fDistance;
	if(fAngle < -1.0f) fAngle = -1.0f;
	if(fAngle > 1.0f) fAngle = 1.0f;
	out.fPhi = (float) acos(fAngle);

	out.fTheta = (float) atan2(in.y, in.x);

	// convert to angles and correct interval
	out.fPhi /= M_PI / 180.0f;
	out.fTheta /= M_PI / 180.0f;

	// only positive rotations
	if(out.fTheta < 0.0)
		out.fTheta = 360.0f + out.fTheta;

	// be aware of limit
	if(out.fPhi > 180.0)
		out.fPhi = out.fPhi - 180.0f;
}

void MathTools::convert(Eigen::Vector3f in, TSphereCoord& out)
{
    Vec3d vec = {in(0), in(1), in(2)};
    convert(vec, out);
}

